// prepreka
// s pogojnimi spremenljivkami
//    Ko nit pride do pthread_cond_wait se ustavi in čaka, da jo druga nit zbudi. 
//	  Ko se ustavi, odklene ključavnico, da dovoli drugim nitim, da vstopajo v kritično sekcijo.
//    Z broadcast damo vsem spečim nitim signal, da naj se zbudijo, takrat se ključavnica spet zaklene.
// Stavek while je priporočljiv, saj lahko tudi kakšen drug dogodek kot broadcast zbudi nit. V tem primeru bo
// funkcija vrnila napako (result !=0 ) in se zaradi while ponovno poklicala.

#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <Windows.h>

#define	NTHREADS	4
#define NPRINTS		5

int shared;

int threads;
pthread_mutex_t lock;
pthread_cond_t	condition;

void* barrier(void* arg) 
{
	int i;
	int myid = (int)arg;	

	for(i=0; i<NPRINTS; i++)
	{
		if( myid == 0 )
			shared++;

		// prepreka - zacetek
		pthread_mutex_lock(&lock);
		threads++;
		if( threads < NTHREADS )
			while( pthread_cond_wait(&condition, &lock) != 0 );
		else
		{
			threads = 0;
			pthread_cond_broadcast(&condition);
		}
		pthread_mutex_unlock(&lock);
		// prepreka - konec

		printf("Nit %d: vrednost: %d\n", myid, shared);
	}

	return NULL;
}

int main(int argc, char *argv[]) 
{
	int i;
	pthread_t t[NTHREADS];

	shared = 0;

	threads = 0;
	pthread_mutex_init(&lock, NULL);
	pthread_cond_init(&condition, NULL);

	for(i=0; i<NTHREADS; i++)
		pthread_create(&t[i], NULL, barrier, (void *)i);

	for(i=0; i<NTHREADS; i++)
		pthread_join(t[i], NULL);

	pthread_mutex_destroy(&lock);
	pthread_cond_destroy(&condition);

	return 0;
}
