// producer - consumer, rešitev s pogojnimi spremenljivkami
// za pravi vrstni red:
//   - PRODUCERS = CONSUMERS = BUFFER_SIZE = 1, 
//   - task_generate takoj pred buffer_put in task_process takoj za buffer_get

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#define PRODUCERS       2
#define CONSUMERS       2
#define BUFFER_SIZE     2
#define TASKS           5

struct task 
{
    int id;
};

int buffer[BUFFER_SIZE];
int buffer_idx_put = 0;
int buffer_idx_get = 0;
int buffer_num;

int task_counter = 0;

pthread_mutex_t m_buffer = PTHREAD_MUTEX_INITIALIZER;	
pthread_cond_t c_rel_prod = PTHREAD_COND_INITIALIZER; 
pthread_cond_t c_rel_cons = PTHREAD_COND_INITIALIZER; 

void task_generate (int producerid, int taskid, struct task *t)
{
    t->id = 10 * producerid + taskid;
	printf ("P:%02d ", t->id);
	fflush (stdout);
}

void task_process (struct task *t)
{
	printf ("C:%02d ", t->id);
	fflush (stdout);
}

void buffer_put (struct task *t)
{
	buffer[buffer_idx_put] = t->id;
	buffer_idx_put = (buffer_idx_put+1) % BUFFER_SIZE;
	buffer_num++;
	printf("b%d ", buffer_num);
}

void buffer_get (struct task *t)
{
	t->id = buffer[buffer_idx_get];
	buffer_idx_get = (buffer_idx_get+1) % BUFFER_SIZE;
	buffer_num--;
	printf("b%d ", buffer_num);
}

void *producer(void *param)
{
	struct task t;

	for (int i = 0; i < TASKS; i++) 
    {
        // generate task
		task_generate (*((int *)param), i, &t);

		// put task to buffer
		pthread_mutex_lock (&m_buffer);
		while (buffer_num == BUFFER_SIZE)
			pthread_cond_wait (&c_rel_prod, &m_buffer);
		buffer_put (&t);
		pthread_cond_signal (&c_rel_cons);
		pthread_mutex_unlock (&m_buffer);
	}
}

void *consumer(void *param)
{
	struct task t;

	while (1) 
    {
		// get task from buffer
		pthread_mutex_lock (&m_buffer);
		while (buffer_num == 0)
			pthread_cond_wait (&c_rel_cons, &m_buffer);
		buffer_get (&t);
		pthread_cond_signal (&c_rel_prod);
		pthread_mutex_unlock (&m_buffer);

        // process task
		task_process (&t);
	}
}

int main (int argc, char *argv[])
{
	pthread_t tp[PRODUCERS];
    pthread_t tc[CONSUMERS];
	int tpid[PRODUCERS];
    
	for (int i=0; i<PRODUCERS; i++)
	{
		tpid[i] = i;
        pthread_create (&tp[i], NULL, producer, &tpid[i]);
	}

	for (int i=0; i<CONSUMERS; i++)
        pthread_create (&tc[i], NULL, consumer, NULL);

    for (int i=0; i<PRODUCERS; i++)
	    pthread_join (tp[i],NULL);

    printf("\n");
}
